/**
 * Copyright (C) 2010-2012 Agrosense info@agrosense.eu
 *
 * Licensed under the Eclipse Public License - v 1.0 (the "License"); you may
 * not use this file except in compliance with the License. You may obtain a
 * copy of the License at
 *
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations under
 * the License.
 */
package nl.cloudfarming.client.task.grid;

import com.vividsolutions.jts.geom.*;
import java.util.List;
import nl.cloudfarming.client.field.model.TreatmentZone;
import org.geotools.geometry.jts.JTSFactoryFinder;

/**
 * Static class to generate a grid from a collection of treatmentzones and a
 * specified amount of columns and rows.
 *
 * @author Gerben Feenstra
 */
public class GridGenerator {

    private static final String EMPTY_TREATMENTZONE = "Empty";

    /**
     * Create a {@link TreatmentZoneGrid} using the passed parameters.
     *
     * Note1: grid y-axis goes from south to north, x-axis goes from west to
     * east. Because y-axis goes from south to north, the order of rows in the
     * returned grid goes from south to north as well. (Position 0 in the List
     * being most southern row in the grid)
     *
     * Note2: TreatmentzoneCode with value 0 is reserved! If no treatmentzone
     * can be found for a gridcell, 0 will be used as the treatmentzoneCode.
     *
     * @param cellWidth the width of the cells used for the grid
     * @param cellHeight the height of the cells used for the grid
     * @param startLongitude the starting longitude of the grid
     * @param startLatitude the starting latitude of the grid
     * @param numberOfRows the number of rows the grid should have
     * @param numberofCols the number of columns the grid should have
     * @param treatmentZones the treatment zones that should be placed in the
     * grid
     * @param matcher the matcher that should be used to determine the geometry
     * for a specific grid cell
     * @return A TreatmentZoneGrid containing all the passed treatment zones
     */
    public static TreatmentZoneGrid generate(double cellWidth, double cellHeight,
            double startLongitude, double startLatitude,
            int numberOfRows, int numberofCols,
            List<TreatmentZone> treatmentZones,
            CellMatchingStrategy matcher){
        // Create empty treatmentzone
        TreatmentZoneGrid grid = new TreatmentZoneGrid(numberOfRows, numberofCols, cellWidth, cellHeight, startLongitude, startLatitude, treatmentZones);

        //latitude needs to be inverted because it starts at the bottom
        startLatitude = startLatitude + (numberOfRows * cellHeight);

        // Row number starting at 0 and working up
        for (int rowNumber = 0; rowNumber < numberOfRows; rowNumber++) {
            // Loop through each column in this row
            for (int columnNumber = 0; columnNumber < numberofCols; columnNumber++) {

                Geometry gridCellGeometry = getCellGeometry(startLongitude, startLatitude, cellWidth, cellHeight, columnNumber, rowNumber);
                TreatmentZone treatmentZone = matcher.findTreatmentZone(treatmentZones, gridCellGeometry);

                if (treatmentZone != null) {
                    grid.insert(rowNumber, columnNumber, grid.getTreatmentZoneCode(treatmentZone), grid.getTreatmentZoneValue(treatmentZone));
                } else {
                    grid.insert(rowNumber, columnNumber, (short) 0, 0);
                }
            }
        }

        return grid;
    }

    /**
     * Create a {@link TreatmentZoneGrid} using the passed parameters.
     *
     * Note1: grid y-axis goes from south to north, x-axis goes from west to
     * east. Because y-axis goes from south to north, the order of rows in the
     * returned grid goes from south to north as well. (Position 0 in the List
     * being most southern row in the grid)
     *
     * Note2: TreatmentzoneCode with value 0 is reserved! If no treatmentzone
     * can be found for a gridcell, 0 will be used as the treatmentzoneCode.
     *
     * @param cellWidth the width of the cells used for the grid
     * @param cellHeight the height of the cells used for the grid
     * @param startLongitude the starting longitude of the grid
     * @param startLatitude the starting latitude of the grid
     * @param numberOfRows the number of rows the grid should have
     * @param numberofCols the number of columns the grid should have
     * @param treatmentZones the treatment zones that should be placed in the
     * grid
     * @return A TreatmentZoneGrid containing all the passed treatment zones
     */
    public static TreatmentZoneGrid generate(double cellWidth, double cellHeight,
            double startLongitude, double startLatitude,
            int numberOfRows, int numberofCols,
            List<TreatmentZone> treatmentZones){
        // use most accurate matcher as default:
        return generate(cellWidth, cellHeight, startLongitude, startLatitude, numberOfRows, numberofCols, treatmentZones, new IntersectionStrategy());
    }

    /**
     * Return a geometry of the specified cell (position), calculated using
     * Startingpoint X, Y as a basis The starting point is the max east and max
     * south position of the containing grid The column and row starting point
     * is assumed to be the top left corner
     *
     * @param startLongitude Startingpoint X
     * @param startLatitude Startingpoint Y
     * @param cellWidth The width of a single cell in the grid
     * @param cellHeight The height of a single cell in the grid
     * @param columnNumber x position in the grid
     * @param rowNumber y position in the grid
     * @return the geometry of the gridcell
     */
    protected static Geometry getCellGeometry(double startLongitude, double startLatitude, double cellWidth, double cellHeight, int columnNumber, int rowNumber) {

        double cellMinX = startLongitude + (columnNumber * cellWidth);
        double cellMinY = startLatitude - ((rowNumber * cellHeight) + cellHeight);
        double cellMaxX = startLongitude + (columnNumber * cellWidth) + cellWidth;
        double cellMaxY = startLatitude - (rowNumber * cellHeight);

        // Create new geometry
        GeometryFactory factory = JTSFactoryFinder.getGeometryFactory(null);
        Coordinate[] coordinates = new Coordinate[5];

        coordinates[0] = new Coordinate(cellMinX, cellMinY);
        coordinates[1] = new Coordinate(cellMinX, cellMaxY);
        coordinates[2] = new Coordinate(cellMaxX, cellMaxY);
        coordinates[3] = new Coordinate(cellMaxX, cellMinY);
        coordinates[4] = new Coordinate(cellMinX, cellMinY);

        LinearRing ring = factory.createLinearRing(coordinates);
        return factory.createPolygon(ring, null);
    }

    public static interface CellMatchingStrategy {

        /**
         * Find the treatmentzone for the specified gridcell. A gridcell can
         * have at most one treatmentzone.
         *
         * @param treatmentZones List of possible treatmentzones
         * @param gridCell Geometry of the gridcell
         * @return the matched treatmentzone, or null when none was found.
         */
        TreatmentZone findTreatmentZone(final List<TreatmentZone> treatmentZones, final Geometry gridCell);
    }

    /**
     * Match a treatmentzone to the cell by testing for intersections. If
     * multiple treatmentzones intersect with the gridcell the treatmentzone
     * which has the largest intersection area with the gridcell will be
     * returned.
     */
    public static class IntersectionStrategy implements CellMatchingStrategy {

        @Override
        public TreatmentZone findTreatmentZone(List<TreatmentZone> treatmentZones, Geometry gridCell) {
            // initial maximum of 0 also prevents selection of geometries that only share a boundary.
            double largestIntersectionArea = 0.0;
            TreatmentZone largestIntersectionTZ = null;

            // Find the treatmentzone which has the largest intersection area with the specified gridcell:
            for (TreatmentZone treatmentZone : treatmentZones) {
                Geometry treatmentZoneGeometry = treatmentZone.getGeometry();
                if (gridCell.intersects(treatmentZoneGeometry)) {
                    double intersectionArea = gridCell.intersection(treatmentZoneGeometry).getArea();
                    if (intersectionArea > largestIntersectionArea) {
                        largestIntersectionArea = intersectionArea;
                        largestIntersectionTZ = treatmentZone;
                    }
                }
            }
            return largestIntersectionTZ;
        }
    }

    /**
     * Match a treatmentzone to the cell by testing if it contains the cells
     * centroid. about twice as fast as the intersection strategy, but less
     * accurate.
     */
    public static class CentroidStrategy implements CellMatchingStrategy {

        @Override
        public TreatmentZone findTreatmentZone(List<TreatmentZone> treatmentZones, Geometry gridCell) {
            // Find the treatmentzone which contains the center of the specified gridcell:
            for (TreatmentZone treatmentZone : treatmentZones) {
                if (treatmentZone.getGeometry().contains(gridCell.getCentroid())) {
                    return treatmentZone;
                }
            }
            return null;
        }
    }
}
