/*
 * Copyright (C) 2010-2012 Agrosense info@agrosense.eu
 * 
 * Licensed under the Eclipse Public License - v 1.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *          http://www.eclipse.org/legal/epl-v10.html
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package nl.cloudfarming.client.task.windows;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Container;
import javax.lang.model.type.NullType;
import javax.swing.*;
import net.miginfocom.swing.MigLayout;
import nl.cloudfarming.client.geoviewer.render.GridCoverage2DPreviewPanel;
import nl.cloudfarming.client.task.ExportableTask;
import nl.cloudfarming.client.task.grid.TreatmentZoneGrid;
import nl.cloudfarming.client.util.swing.ASBodyPanel;
import org.netbeans.api.progress.ProgressHandle;
import org.netbeans.api.progress.ProgressHandleFactory;
import org.netbeans.core.spi.multiview.CloseOperationState;
import org.netbeans.core.spi.multiview.MultiViewElement;
import org.netbeans.core.spi.multiview.MultiViewElementCallback;
import org.openide.awt.UndoRedo;
import org.openide.util.Lookup;
import org.openide.util.NbBundle;
import org.openide.util.NbBundle.Messages;
import org.openide.windows.TopComponent;

/**
 *
 * @author Maciek Dulkiewicz
 */
@MultiViewElement.Registration(displayName = "#ExportableTaskPreviewElement_display_name",
iconBase = "nl/cloudfarming/client/task/to_do_list_checked_1_16.png",
mimeType = ExportableTask.MIME_TYPE,
persistenceType = TopComponent.PERSISTENCE_NEVER,
preferredID = "ExportableTaskPreview",
position = 200)
@Messages({
    "ExportableTaskPreviewElement_display_name=Grid preview",
    "ExportableTaskPreviewElement_loading_panel_label=Loading grid"
})
public final class ExportableTaskPreviewElement extends TopComponent implements MultiViewElement {

    private JToolBar toolbar = new JToolBar();
    private transient MultiViewElementCallback callback;
    private Lookup lookup;
    private ASBodyPanel bodyPanel;
    private ExportableTask task;
    private JPanel contentPanel;
    private static final Color BACKGROUND_COLOR = Color.WHITE;

    public ExportableTaskPreviewElement(Lookup lookup) {
        initComponents();
        this.lookup = lookup;
        task = lookup.lookup(ExportableTask.class);
        setName(Bundle.ExportableTaskPreviewElement_display_name());
    }

    /**
     * This method is responsible for making sure an initial grid is generated
     * by the task in a background thread. It calls the needed methods in the
     * {@link ExportableTask} using a SwingWorker and lets the null null null
     * null null null null null null null null null null     {@link ExportableTaskPreviewElement#replaceLoadingPanel(nl.cloudfarming.client.task.grid.TreatmentZoneGrid)
     * method replace the loadingPanel with an actual grid preview.
     */
    private void initGridPreview() {
        new GridConstructor().execute();
    }

    private class GridConstructor extends SwingWorker<NullType, Object> {

        private ProgressHandle progressHandle;
        private TreatmentZoneGrid grid;

        public GridConstructor() {
            contentPanel.removeAll();
            JPanel loadingPanel = new JPanel(new MigLayout("wrap", "[center, grow]", "[center, grow]"));
            progressHandle = ProgressHandleFactory.createHandle(Bundle.ExportableTaskPreviewElement_loading_panel_label());
            JComponent progressComponent = ProgressHandleFactory.createProgressComponent(progressHandle);
            loadingPanel.add(ProgressHandleFactory.createMainLabelComponent(progressHandle));
            loadingPanel.add(progressComponent);
            loadingPanel.setBackground(BACKGROUND_COLOR);
            contentPanel.add(loadingPanel);
        }

        @Override
        protected NullType doInBackground() throws Exception {
            progressHandle.start();
            ExportableTaskError[] errors = task.validate();
            if (errors.length > 0) {
                ExportableTaskValidationMessagePanel panel = new ExportableTaskValidationMessagePanel();
                for (ExportableTaskError error : errors) {
                    panel.addValidationMessage(error.getMessage());
                }
                addPanel(panel);
            } else {
                grid = task.getExportableGrid();
            }
            progressHandle.finish();
            return null;
        }

        /**
         * This method constructs a {@link GridCoverage2DPreviewPanel} using the
         * generated grid. It then removes all the components from it's panel
         * and replaces it with a newly constructed  {@link GridCoverage2DPreviewPanel}
         *
         */
        @Override
        protected void done() {
            if (grid != null) {
                GridCoverage2DPreviewPanel gridPanel = new GridCoverage2DPreviewPanel(grid.getGridCoverage2D());
                addPanel(gridPanel);
            }
        }

        /**
         * Add a panel to the contentpanel
         *
         * @param panel
         */
        private void addPanel(Container panel) {
            contentPanel.removeAll();
            contentPanel.add(panel);
            contentPanel.revalidate();
        }

        /**
         * Panel which is displayed when one ore more values of a treatment zone
         * is null
         */
        @Messages({"ExportableTaskValidationMessagePanel error message=Cannot generate grid due to: "})
        private class ExportableTaskValidationMessagePanel extends JPanel {

            public ExportableTaskValidationMessagePanel() {
                this.setLayout(new MigLayout("wrap"));
                this.setBackground(BACKGROUND_COLOR);
                JLabel message = new JLabel(Bundle.ExportableTaskValidationMessagePanel_error_message());
                this.add(message);
            }

            public void addValidationMessage(String message) {
                this.add(new JLabel(message));
            }
        }
    }

    /**
     * An enum of error messages for all the possible validation issues which a
     * task can have
     *
     * @author Merijn Zengers <Merijn.Zengers@gmail.nl>
     */
    @NbBundle.Messages({"ExportableTaskErrorEnum treatment zone null value error message=Task is invalid. One ore more treatment zone values are empty"})
    public enum ExportableTaskError {

        TREATMENT_ZONE_NULL_VALUE(Bundle.ExportableTaskErrorEnum_treatment_zone_null_value_error_message());
        private String message;

        private ExportableTaskError(String message) {
            this.message = message;
        }

        public String getMessage() {
            return message;
        }
    }

    /**
     * Returns the initial panel containing another panel displaying a simple
     * loading message. This panel is displayed while the
     * {@link TreatmentZoneGrid} for the task is generated.
     *
     * @return ASBodyPanel containing a simple loading panel
     */
    @Override
    public JComponent getVisualRepresentation() {
        if (bodyPanel == null) {
            contentPanel = new JPanel();
            contentPanel.setLayout(new BorderLayout());
            bodyPanel = new ASBodyPanel(contentPanel);

        }
        return bodyPanel;
    }

    /**
     * This method is called from within the constructor to initialize the form.
     * WARNING: Do NOT modify this code. The content of this method is always
     * regenerated by the Form Editor.
     */
    // <editor-fold defaultstate="collapsed" desc="Generated Code">//GEN-BEGIN:initComponents
    private void initComponents() {

        javax.swing.GroupLayout layout = new javax.swing.GroupLayout(this);
        this.setLayout(layout);
        layout.setHorizontalGroup(
            layout.createParallelGroup(javax.swing.GroupLayout.Alignment.LEADING)
            .addGap(0, 400, Short.MAX_VALUE)
        );
        layout.setVerticalGroup(
            layout.createParallelGroup(javax.swing.GroupLayout.Alignment.LEADING)
            .addGap(0, 300, Short.MAX_VALUE)
        );
    }// </editor-fold>//GEN-END:initComponents

    // Variables declaration - do not modify//GEN-BEGIN:variables
    // End of variables declaration//GEN-END:variables
    /**
     * This method is called as soon as the ExportableTaskPreviewElement has
     * been opened and the panel has been filled with the default loadingPanel.
     * This method makes sure the initial version of the grid preview is
     * generated by the task.
     */
    @Override
    public void componentOpened() {
    }

    @Override
    public void componentClosed() {
    }

    void writeProperties(java.util.Properties p) {
        // better to version settings since initial version as advocated at
        // http://wiki.apidesign.org/wiki/PropertyFiles
        p.setProperty("version", "1.0");
        // TODO store your settings
    }

    void readProperties(java.util.Properties p) {
        String version = p.getProperty("version");
    }

    @Override
    public JComponent getToolbarRepresentation() {
        return toolbar;
    }

    @Override
    public Action[] getActions() {
        return new Action[0];
    }

    @Override
    public Lookup getLookup() {
        return lookup;
    }

    @Override
    public void componentShowing() {
    }

    @Override
    public void componentHidden() {
    }

    @Override
    public void componentActivated() {
        initGridPreview();
    }

    @Override
    public void componentDeactivated() {
    }

    @Override
    public UndoRedo getUndoRedo() {
        return UndoRedo.NONE;
    }

    @Override
    public void setMultiViewCallback(MultiViewElementCallback callback) {
        this.callback = callback;
    }

    @Override
    public CloseOperationState canCloseElement() {
        return CloseOperationState.STATE_OK;
    }

    @Override
    public int getPersistenceType() {
        return PERSISTENCE_NEVER;
    }
}
